@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -89,7 +89,7 @@ static void list_emit(void)
     if (!listlinep && !listdata[0])
         return;
 
-    fprintf(listfp, "%6"PRId32" ", ++listlineno);
+    fprintf(listfp, "%6"PRId32" ", listlineno);
 
     if (listdata[0])
         fprintf(listfp, "%08"PRIX32" %-*s", listoffset, LIST_HEXBIT + 1,
@@ -111,7 +111,7 @@ static void list_emit(void)
     listdata[0] = '\0';
 
     if (listerror[0]) {
-	fprintf(listfp, "%6"PRId32"          ", ++listlineno);
+	fprintf(listfp, "%6"PRId32"          ", listlineno);
 	for (i = 0; i < LIST_HEXBIT; i++)
 	    putc('*', listfp);
 	
@@ -126,12 +126,17 @@ static void list_emit(void)
     }
 }
 
-static void list_init(char *fname, efunc error)
+static void list_init(const char *fname)
 {
+    if (!fname || fname[0] == '\0') {
+	listfp = NULL;
+	return;
+    }
+
     listfp = fopen(fname, "w");
     if (!listfp) {
-        error(ERR_NONFATAL, "unable to open listing file `%s'",
-	      fname);
+	nasm_error(ERR_NONFATAL, "unable to open listing file `%s'",
+		   fname);
         return;
     }
 
@@ -197,7 +202,7 @@ static void list_output(int32_t offset, const void *data,
 {
     char q[20];
 
-    if (!listp || suppress || user_nolist)      /* fbk - 9/2/00 */
+    if (!listp || suppress || user_nolist)
         return;
 
     switch (type) {
@@ -216,7 +221,7 @@ static void list_output(int32_t offset, const void *data,
 	break;
     }
     case OUT_ADDRESS:
-	list_address(offset, "[]", *(int64_t *)data, abs(size));
+      list_address(offset, "[]", *(int64_t *)data, abs((int)size));
 	break;
     case OUT_REL1ADR:
 	list_address(offset, "()", *(int64_t *)data, 1);
@@ -243,10 +248,9 @@ static void list_line(int type, char *line)
 {
     if (!listp)
         return;
-    if (user_nolist) {          /* fbk - 9/2/00 */
-        listlineno++;
-        return;
-    }
+
+    if (user_nolist)
+      return;
 
     if (mistack && mistack->inhibiting) {
         if (type == LIST_MACRO)
@@ -258,6 +262,7 @@ static void list_line(int type, char *line)
         }
     }
     list_emit();
+    listlineno = src_get_linnum();
     listlinep = true;
     strncpy(listline, line, LIST_MAX_LEN - 1);
     listline[LIST_MAX_LEN - 1] = '\0';
@@ -321,7 +326,7 @@ static void list_error(int severity, const char *pfx, const char *msg)
 }
 
 
-ListGen nasmlist = {
+static const struct lfmt nasm_list = {
     list_init,
     list_cleanup,
     list_output,
@@ -330,3 +335,5 @@ ListGen nasmlist = {
     list_downlevel,
     list_error
 };
+
+const struct lfmt *lfmt = &nasm_list;
