@@ -44,7 +44,9 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "preproc.h"
 #include "saa.h"
+#include "hashtbl.h"
 #include "output/outlib.h"
 #include "output/pecoff.h"
 #include "md5.h"
@@ -73,8 +75,21 @@ struct dfmt df_cv8 = {
 /*******************************************************************************
  * dfmt callbacks
  ******************************************************************************/
+struct source_file;
+
 struct source_file {
-    char *name;
+    const char *filename;
+    char *fullname;
+    uint32_t fullnamelen;
+
+    struct source_file *next;
+
+    uint32_t filetbl_off;
+    uint32_t sourcetbl_off;
+
+    struct SAA *lines;
+    uint32_t num_lines;
+
     unsigned char md5sum[MD5_HASHBYTES];
 };
 
@@ -123,10 +138,15 @@ struct cv8_state {
 
     uint32_t text_offset;
 
-    struct source_file source_file;
+    struct source_file *source_files, **source_files_tail;
+    const char *last_filename;
+    struct source_file *last_source_file;
+    struct hash_table file_hash;
+    unsigned num_files;
+    uint32_t total_filename_len;
 
-    struct SAA *lines;
-    uint32_t num_lines;
+
+    unsigned total_lines;
 
     struct SAA *symbols;
     struct cv8_symbol *last_sym;
@@ -134,7 +154,10 @@ struct cv8_state {
     unsigned symbol_lengths;
     unsigned total_syms;
 
-    char *cwd;
+    struct {
+        char *name;
+        size_t namebytes;
+    } outfile;
 };
 struct cv8_state cv8_state;
 
@@ -150,16 +173,20 @@ static void cv8_init(void)
 
     cv8_state.text_offset = 0;
 
-    cv8_state.lines = saa_init(sizeof(struct linepair));
-    cv8_state.num_lines = 0;
+    cv8_state.source_files = NULL;
+    cv8_state.source_files_tail = &cv8_state.source_files;
+    hash_init(&cv8_state.file_hash, HASH_MEDIUM);
+
+    cv8_state.num_files = 0;
+    cv8_state.total_filename_len = 0;
+
+    cv8_state.total_lines = 0;
 
     cv8_state.symbols = saa_init(sizeof(struct cv8_symbol));
     cv8_state.last_sym = NULL;
-
-    cv8_state.cwd = nasm_realpath(".");
 }
 
-static void register_file(const char *filename);
+static struct source_file *register_file(const char *filename);
 static struct coff_Section *find_section(int32_t segto);
 
 static void cv8_linenum(const char *filename, int32_t linenumber,
@@ -167,9 +194,9 @@ static void cv8_linenum(const char *filename, int32_t linenumber,
 {
     struct coff_Section *s;
     struct linepair *li;
+    struct source_file *file;
 
-    if (cv8_state.source_file.name == NULL)
-        register_file(filename);
+    file = register_file(filename);
 
     s = find_section(segto);
     if (s == NULL)
@@ -178,11 +205,12 @@ static void cv8_linenum(const char *filename, int32_t linenumber,
     if ((s->flags & IMAGE_SCN_MEM_EXECUTE) == 0)
         return;
 
-    li = saa_wstruct(cv8_state.lines);
+    li = saa_wstruct(file->lines);
     li->file_offset = cv8_state.text_offset;
     li->linenumber = linenumber;
 
-    cv8_state.num_lines++;
+    file->num_lines++;
+    cv8_state.total_lines++;
 }
 
 static void cv8_deflabel(char *name, int32_t segment, int64_t offset,
@@ -276,25 +304,30 @@ static void build_type_table(struct coff_Section *const sect);
 static void cv8_cleanup(void)
 {
     struct cv8_symbol *sym;
+    struct source_file *file;
 
     struct coff_Section *symbol_sect = coff_sects[cv8_state.symbol_sect];
     struct coff_Section *type_sect = coff_sects[cv8_state.type_sect];
 
+    cv8_state.outfile.name = nasm_realpath(coff_outfile);
+    cv8_state.outfile.namebytes = strlen(cv8_state.outfile.name) + 1;
+
     build_symbol_table(symbol_sect);
     build_type_table(type_sect);
 
-    if (cv8_state.source_file.name != NULL)
-        free(cv8_state.source_file.name);
-
-    if (cv8_state.cwd != NULL)
-        free(cv8_state.cwd);
-
-    saa_free(cv8_state.lines);
+    list_for_each(file, cv8_state.source_files) {
+        nasm_free(file->fullname);
+        saa_free(file->lines);
+        free(file);
+    }
+    hash_free(&cv8_state.file_hash);
 
     saa_rewind(cv8_state.symbols);
     while ((sym = saa_rstruct(cv8_state.symbols)))
-        free(sym->name);
+        nasm_free(sym->name);
     saa_free(cv8_state.symbols);
+
+    nasm_free(cv8_state.outfile.name);
 }
 
 /*******************************************************************************
@@ -308,7 +341,7 @@ static void calc_md5(const char *const filename,
     FILE *f;
     MD5_CTX ctx;
 
-    f = fopen(filename, "r");
+    f = pp_input_fopen(filename, "rb");
     if (!f)
         goto done;
 
@@ -338,11 +371,50 @@ done:
     return;
 }
 
-static void register_file(const char *filename)
+static struct source_file *register_file(const char *filename)
 {
-    cv8_state.source_file.name = nasm_realpath(filename);
-    memset(cv8_state.source_file.md5sum, 0, MD5_HASHBYTES);
-    calc_md5(filename, cv8_state.source_file.md5sum);
+    struct source_file *file;
+    void **filep;
+    char *fullpath;
+    struct hash_insert hi;
+
+    /*
+     * The common case is that we are invoked with the same filename
+     * as we were last time.  Make this a pointer comparison: this is
+     * safe because the NASM core code allocates each filename once
+     * and never frees it.
+     */
+    if (likely(cv8_state.last_filename == filename))
+        return cv8_state.last_source_file;
+
+    cv8_state.last_filename = filename;
+
+    filep = hash_find(&cv8_state.file_hash, filename, &hi);
+    if (likely(filep)) {
+        file = *filep;
+    } else {
+        /* New filename encounter */
+
+        fullpath = nasm_realpath(filename);
+
+        file = nasm_zalloc(sizeof(*file));
+
+        file->filename = filename;
+        file->fullname = fullpath;
+        file->fullnamelen = strlen(fullpath);
+        file->lines = saa_init(sizeof(struct linepair));
+        *cv8_state.source_files_tail = file;
+        cv8_state.source_files_tail = &file->next;
+        calc_md5(fullpath, file->md5sum);
+
+        hash_add(&hi, filename, file);
+
+        cv8_state.num_files++;
+        cv8_state.total_filename_len += file->fullnamelen + 1;
+    }
+
+    cv8_state.last_source_file = file;
+    return file;
 }
 
 static struct coff_Section *find_section(int32_t segto)
@@ -432,40 +504,63 @@ static inline void section_wbytes(struct coff_Section *sect, const void *buf,
 
 static void write_filename_table(struct coff_Section *const sect)
 {
-    uint32_t field_length = 0;
-    size_t filename_len = strlen(cv8_state.source_file.name);
+    uint32_t field_length;
+    uint32_t tbl_off = 1;    /* offset starts at 1 to skip NULL entry */
+    struct source_file *file;
 
-    field_length = 1 + filename_len + 1;
+    nasm_assert(cv8_state.source_files != NULL);
+    nasm_assert(cv8_state.num_files > 0);
+    nasm_assert(cv8_state.total_filename_len > 0);
+
+    field_length = 1 + cv8_state.total_filename_len;
 
     section_write32(sect, 0x000000F3);
     section_write32(sect, field_length);
 
     section_write8(sect, 0);
-    section_wbytes(sect, cv8_state.source_file.name, filename_len + 1);
+
+    list_for_each(file, cv8_state.source_files) {
+        section_wbytes(sect, file->fullname, file->fullnamelen + 1);
+        file->filetbl_off = tbl_off;
+        tbl_off += file->fullnamelen + 1;
+    }
 }
 
 static void write_sourcefile_table(struct coff_Section *const sect)
 {
-    uint32_t field_length = 0;
+    const uint32_t entry_size = 4 + 2 + MD5_HASHBYTES + 2;
 
-    field_length = 4 + 2 + MD5_HASHBYTES + 2;
+    uint32_t field_length = 0;
+    uint32_t tbl_off = 0;
+    struct source_file *file;
+
+    field_length = entry_size * cv8_state.num_files;
 
     section_write32(sect, 0x000000F4);
     section_write32(sect, field_length);
 
-    section_write32(sect, 1); /* offset of filename in filename str table */
-    section_write16(sect, 0x0110);
-    section_wbytes(sect, cv8_state.source_file.md5sum, MD5_HASHBYTES);
-    section_write16(sect, 0);
+    list_for_each(file, cv8_state.source_files) {
+        nasm_assert(file->filetbl_off > 0);
+        section_write32(sect, file->filetbl_off);
+        section_write16(sect, 0x0110);
+        section_wbytes(sect, file->md5sum, MD5_HASHBYTES);
+        section_write16(sect, 0);
+
+        file->sourcetbl_off = tbl_off;
+        tbl_off += entry_size;
+    }
 }
 
 static void write_linenumber_table(struct coff_Section *const sect)
 {
+    const uint32_t file_field_len = 12;
+    const uint32_t line_field_len = 8;
+
     int i;
     uint32_t field_length = 0;
     size_t field_base;
+    struct source_file *file;
     struct coff_Section *s;
-    struct linepair *li;
 
     for (i = 0; i < coff_nsects; i++) {
         if (!strncmp(coff_sects[i]->name, ".text", 5))
@@ -476,7 +571,9 @@ static void write_linenumber_table(struct coff_Section *const sect)
         return;
     s = coff_sects[i];
 
-    field_length = 12 + 12 + (cv8_state.num_lines * 8);
+    field_length = 12;
+    field_length += (cv8_state.num_files * file_field_len);
+    field_length += (cv8_state.total_lines * line_field_len);
 
     section_write32(sect, 0x000000F2);
     section_write32(sect, field_length);
@@ -493,32 +590,33 @@ static void write_linenumber_table(struct coff_Section *const sect)
     register_reloc(sect, ".text", field_base + 4,
         win64 ? IMAGE_REL_AMD64_SECTION : IMAGE_REL_I386_SECTION);
 
-    /* 1 or more source mappings (we assume only 1) */
-    section_write32(sect, 0);
-    section_write32(sect, cv8_state.num_lines);
-    section_write32(sect, 12 + (cv8_state.num_lines * 8));
+    list_for_each(file, cv8_state.source_files) {
+        struct linepair *li;
 
-    /* the pairs */
-    saa_rewind(cv8_state.lines);
-    while ((li = saa_rstruct(cv8_state.lines))) {
-        section_write32(sect, li->file_offset);
-        section_write32(sect, li->linenumber |= 0x80000000);
+        /* source mapping */
+        section_write32(sect, file->sourcetbl_off);
+        section_write32(sect, file->num_lines);
+        section_write32(sect, file_field_len + (file->num_lines * line_field_len));
+
+        /* the pairs */
+        saa_rewind(file->lines);
+        while ((li = saa_rstruct(file->lines))) {
+            section_write32(sect, li->file_offset);
+            section_write32(sect, li->linenumber |= 0x80000000);
+        }
     }
 }
 
-static uint16_t write_symbolinfo_obj(struct coff_Section *sect,
-        const char sep)
+static uint16_t write_symbolinfo_obj(struct coff_Section *sect)
 {
     uint16_t obj_len;
 
-    obj_len = 2 + 4 + strlen(cv8_state.cwd)+ 1 + strlen(coff_outfile) +1;
+    obj_len = 2 + 4 + cv8_state.outfile.namebytes;
 
     section_write16(sect, obj_len);
     section_write16(sect, 0x1101);
     section_write32(sect, 0); /* ASM language */
-    section_wbytes(sect, cv8_state.cwd, strlen(cv8_state.cwd));
-    section_write8(sect, sep);
-    section_wbytes(sect, coff_outfile, strlen(coff_outfile)+1);
+    section_wbytes(sect, cv8_state.outfile.name, cv8_state.outfile.namebytes);
 
     return obj_len;
 }
@@ -606,15 +704,14 @@ static uint16_t write_symbolinfo_symbols(struct coff_Section *sect)
 
 static void write_symbolinfo_table(struct coff_Section *const sect)
 {
-    const char sep = '\\';
-    const char *creator_str = "The Netwide Assembler " NASM_VER;
-
-
+    static const char creator_str[] = "The Netwide Assembler " NASM_VER;
     uint16_t obj_length, creator_length, sym_length;
     uint32_t field_length = 0, out_len;
 
-    /* signature, language, workingdir / coff_outfile NULL */
-    obj_length = 2 + 4 + strlen(cv8_state.cwd)+ 1 + strlen(coff_outfile) +1;
+    nasm_assert(cv8_state.outfile.namebytes);
+
+    /* signature, language, outfile NULL */
+    obj_length = 2 + 4 + cv8_state.outfile.namebytes;
     creator_length = 2 + 4 + 4 + 4 + 4 + strlen(creator_str)+1 + 2;
 
     sym_length =    ( cv8_state.num_syms[SYMTYPE_CODE] *  7) +
@@ -632,7 +729,7 @@ static void write_symbolinfo_table(struct coff_Section *const sect)
 
     /* for sub fields, length preceeds type */
 
-    out_len = write_symbolinfo_obj(sect, sep);
+    out_len = write_symbolinfo_obj(sect);
     nasm_assert(out_len == obj_length);
 
     out_len = write_symbolinfo_properties(sect, creator_str);
@@ -701,5 +798,3 @@ static void build_type_table(struct coff_Section *const sect)
         section_write32(sect, 0); /*num params */
     }
 }
-
-
