@@ -223,13 +223,12 @@ static void coff_gen_init(void)
     def_seg = seg_alloc();
 }
 
-static void coff_cleanup(int debuginfo)
+static void coff_cleanup(void)
 {
     struct coff_Reloc *r;
     int i;
 
-    if (debuginfo && ofmt->current_dfmt->cleanup)
-        ofmt->current_dfmt->cleanup();
+    dfmt->cleanup();
 
     coff_write();
     for (i = 0; i < coff_nsects; i++) {
@@ -583,7 +582,7 @@ static void coff_out(int32_t segto, const void *data,
     if (!s) {
         int tempint;            /* ignored */
         if (segto != coff_section_names(".text", 2, &tempint))
-            nasm_error(ERR_PANIC, "strange segment conditions in COFF driver");
+            nasm_panic(0, "strange segment conditions in COFF driver");
         else
             s = coff_sects[coff_nsects - 1];
     }
@@ -603,7 +602,7 @@ static void coff_out(int32_t segto, const void *data,
 
     memset(mydata, 0, sizeof(mydata));
 
-    if (ofmt->current_dfmt && ofmt->current_dfmt->debug_output) {
+    if (dfmt && dfmt->debug_output) {
         struct coff_DebugInfo dinfo;
         dinfo.segto = segto;
         dinfo.seg = segment;
@@ -614,7 +613,7 @@ static void coff_out(int32_t segto, const void *data,
         else
             dinfo.size = realsize(type, size);
 
-        ofmt->current_dfmt->debug_output(type, &dinfo);
+        dfmt->debug_output(type, &dinfo);
     }
 
     if (type == OUT_RESERVE) {
@@ -626,7 +625,7 @@ static void coff_out(int32_t segto, const void *data,
             s->len += size;
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
-            nasm_error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
+            nasm_panic(0, "OUT_RAWDATA with other than NO_SEG");
         coff_sect_write(s, data, size);
     } else if (type == OUT_ADDRESS) {
         int asize = abs((int)size);
@@ -674,7 +673,7 @@ static void coff_out(int32_t segto, const void *data,
               " relocations");
     } else if (type == OUT_REL4ADR) {
         if (segment == segto && !(win64))  /* Acceptable for RIP-relative */
-            nasm_error(ERR_PANIC, "intra-segment OUT_REL4ADR");
+            nasm_panic(0, "intra-segment OUT_REL4ADR");
         else if (segment == NO_SEG && win32)
             nasm_error(ERR_NONFATAL, "Win32 COFF does not correctly support"
                   " relative references to absolute addresses");
@@ -875,7 +874,7 @@ static inline void coff_adjust_relocs(struct coff_Section *s)
     else
     {
         if (ofmt == &of_coff)
-            nasm_error(ERR_FATAL,
+            nasm_fatal(0,
                        "Too many relocations (%d) for section `%s'",
                        s->nrelocs, s->name);
     }
