@@ -63,10 +63,9 @@ static void dbg_init(void)
     fprintf(ofile, "NASM Output format debug dump\n");
 }
 
-static void dbg_cleanup(int debuginfo)
+static void dbg_cleanup(void)
 {
-    (void)debuginfo;
-    of_dbg.current_dfmt->cleanup();
+    dfmt->cleanup();
     while (dbgsect) {
         struct Section *tmp = dbgsect;
         dbgsect = dbgsect->next;
@@ -264,6 +263,7 @@ struct ofmt of_dbg = {
     "Trace of all info passed to output stage",
     "dbg",
     OFMT_TEXT,
+    64,
     debug_debug_arr,
     &debug_debug_form,
     NULL,
