@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2013 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -159,8 +159,6 @@ static struct elf_symbol *lastsym;
 
 /* common debugging routines */
 static void debug32_typevalue(int32_t);
-static void debug32_deflabel(char *, int32_t, int64_t, int, char *);
-static void debug32_directive(const char *, const char *);
 
 /* stabs debugging routines */
 static void stabs32_linenum(const char *filename, int32_t linenumber, int32_t);
@@ -219,13 +217,11 @@ static void elf_init(void)
     def_seg = seg_alloc();
 }
 
-static void elf_cleanup(int debuginfo)
+static void elf_cleanup(void)
 {
     struct elf_reloc *r;
     int i;
 
-    (void)debuginfo;
-
     elf_write();
     for (i = 0; i < nsects; i++) {
         if (sects[i]->type != SHT_NOBITS)
@@ -242,9 +238,7 @@ static void elf_cleanup(int debuginfo)
     saa_free(syms);
     raa_free(bsym);
     saa_free(strs);
-    if (of_elf32.current_dfmt) {
-        of_elf32.current_dfmt->cleanup();
-    }
+    dfmt->cleanup();
 }
 
 static void add_sectname(char *firsthalf, char *secondhalf)
@@ -383,7 +377,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                 stdscan_reset();
                 stdscan_set(p);
                 tokval.t_type = TOKEN_INVALID;
-                e = evaluate(stdscan, NULL, &tokval, NULL, 1, nasm_error, NULL);
+                e = evaluate(stdscan, NULL, &tokval, NULL, 1, NULL);
                 if (e) {
                     if (!is_simple(e))
                         nasm_error(ERR_NONFATAL, "cannot use relocatable"
@@ -422,7 +416,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
             /* we have to be sure at least text section is there */
             int tempint;
             if (segment != elf_section_names(".text", 2, &tempint))
-                nasm_error(ERR_PANIC, "strange segment conditions in ELF driver");
+                nasm_panic(0, "strange segment conditions in ELF driver");
         }
         for (i = 0; i < nsects; i++) {
             if (segment == sects[i]->index) {
@@ -525,8 +519,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                     stdscan_reset();
                     stdscan_set(special + n);
                     tokval.t_type = TOKEN_INVALID;
-                    e = evaluate(stdscan, NULL, &tokval, &fwd, 0, nasm_error,
-                                 NULL);
+		    e = evaluate(stdscan, NULL, &tokval, &fwd, 0, NULL);
                     if (fwd) {
                         sym->nextfwd = fwds;
                         fwds = sym;
@@ -686,7 +679,7 @@ static void elf_out(int32_t segto, const void *data,
     if (!s) {
         int tempint;            /* ignored */
         if (segto != elf_section_names(".text", 2, &tempint))
-            nasm_error(ERR_PANIC, "strange segment conditions in ELF driver");
+            nasm_panic(0, "strange segment conditions in ELF driver");
         else {
             s = sects[nsects - 1];
             i = nsects - 1;
@@ -694,12 +687,10 @@ static void elf_out(int32_t segto, const void *data,
     }
 
     /* again some stabs debugging stuff */
-    if (of_elf32.current_dfmt) {
-        sinfo.offset = s->len;
-        sinfo.section = i;
-        sinfo.name = s->name;
-        of_elf32.current_dfmt->debug_output(TY_STABSSYMLIN, &sinfo);
-    }
+    sinfo.offset = s->len;
+    sinfo.section = i;
+    sinfo.name = s->name;
+    dfmt->debug_output(TY_STABSSYMLIN, &sinfo);
     /* end of debugging stuff */
 
     if (s->type == SHT_NOBITS && type != OUT_RESERVE) {
@@ -709,6 +700,8 @@ static void elf_out(int32_t segto, const void *data,
         return;
     }
 
+    memset(mydata, 0, sizeof(mydata));
+
     switch (type) {
     case OUT_RESERVE:
         if (s->type == SHT_PROGBITS) {
@@ -721,14 +714,14 @@ static void elf_out(int32_t segto, const void *data,
 
     case OUT_RAWDATA:
         if (segment != NO_SEG)
-            nasm_error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
+            nasm_panic(0, "OUT_RAWDATA with other than NO_SEG");
         elf_sect_write(s, data, size);
 	break;
 
     case OUT_ADDRESS:
     {
         bool gnu16 = false;
-        int asize = abs(size);
+        int asize = abs((int)size);
         addr = *(int64_t *)data;
         if (segment != NO_SEG) {
             if (segment % 2) {
@@ -844,7 +837,7 @@ static void elf_out(int32_t segto, const void *data,
 
     case OUT_REL4ADR:
         if (segment == segto)
-            nasm_error(ERR_PANIC, "intra-segment OUT_REL4ADR");
+            nasm_panic(0, "intra-segment OUT_REL4ADR");
         if (segment != NO_SEG && segment % 2) {
             nasm_error(ERR_NONFATAL, "ELF format does not support"
                   " segment base references");
@@ -895,9 +888,9 @@ static void elf_write(void)
      * relocation sections for the user sections.
      */
     nsections = sec_numspecial + 1;
-    if (of_elf32.current_dfmt == &df_stabs)
+    if (dfmt == &df_stabs)
         nsections += 3;
-    else if (of_elf32.current_dfmt == &df_dwarf)
+    else if (dfmt == &df_dwarf)
         nsections += 10;
 
     add_sectname("", ".shstrtab");
@@ -911,12 +904,12 @@ static void elf_write(void)
         }
     }
 
-    if (of_elf32.current_dfmt == &df_stabs) {
+    if (dfmt == &df_stabs) {
         /* in case the debug information is wanted, just add these three sections... */
         add_sectname("", ".stab");
         add_sectname("", ".stabstr");
         add_sectname(".rel", ".stab");
-    } else if (of_elf32.current_dfmt == &df_dwarf) {
+    } else if (dfmt == &df_dwarf) {
         /* the dwarf debug standard specifies the following ten sections,
            not all of which are currently implemented,
            although all of them are defined. */
@@ -1013,7 +1006,7 @@ static void elf_write(void)
             p += strlen(p) + 1;
         }
 
-    if (of_elf32.current_dfmt == &df_stabs) {
+    if (dfmt == &df_stabs) {
         /* for debugging information, create the last three sections
            which are the .stab , .stabstr and .rel.stab sections respectively */
 
@@ -1034,7 +1027,7 @@ static void elf_write(void)
                                stabrellen, sec_symtab, sec_stab, 4, 8);
             p += strlen(p) + 1;
         }
-    } else if (of_elf32.current_dfmt == &df_dwarf) {
+    } else if (dfmt == &df_dwarf) {
             /* for dwarf debugging information, create the ten dwarf sections */
 
             /* this function call creates the dwarf sections in memory */
@@ -1160,8 +1153,8 @@ static struct SAA *elf_build_symtab(int32_t *len, int32_t *local)
       * dwarf needs symbols for debug sections
       * which are relocation targets.
       */
-//*** fix for 32 bit
-     if (of_elf32.current_dfmt == &df_dwarf) {
+    /*** fix for 32 bit ***/
+     if (dfmt == &df_dwarf) {
         dwarf_infosym = *local;
         p = entry;
         WRITELONG(p, 0);        /* no symbol name */
@@ -1340,8 +1333,8 @@ static struct dfmt df_dwarf = {
     "dwarf",
     dwarf32_init,
     dwarf32_linenum,
-    debug32_deflabel,
-    debug32_directive,
+    null_debug_deflabel,
+    null_debug_directive,
     debug32_typevalue,
     dwarf32_output,
     dwarf32_cleanup
@@ -1351,8 +1344,8 @@ static struct dfmt df_stabs = {
     "stabs",
     null_debug_init,
     stabs32_linenum,
-    debug32_deflabel,
-    debug32_directive,
+    null_debug_deflabel,
+    null_debug_directive,
     debug32_typevalue,
     stabs32_output,
     stabs32_cleanup
@@ -1364,6 +1357,7 @@ struct ofmt of_elf32 = {
     "ELF32 (i386) object files (e.g. Linux)",
     "elf32",
     0,
+    32,
     elf32_debugs_arr,
     &df_stabs,
     elf_stdmac,
@@ -1407,22 +1401,6 @@ static void stabs32_linenum(const char *filename, int32_t linenumber,
     currentline = linenumber;
 }
 
-static void debug32_deflabel(char *name, int32_t segment, int64_t offset, int is_global,
-                    char *special)
-{
-   (void)name;
-   (void)segment;
-   (void)offset;
-   (void)is_global;
-   (void)special;
-}
-
-static void debug32_directive(const char *directive, const char *params)
-{
-   (void)directive;
-   (void)params;
-}
-
 static void debug32_typevalue(int32_t type)
 {
     int32_t stype, ssize;
@@ -1593,7 +1571,7 @@ static void stabs32_generate(void)
          * the source-file, the n_desc field should be set to the number
          * of remaining stabs
          */
-        WRITE_STAB(sptr, fileidx[0], 0, 0, 0, strlen(allfiles[0] + 12));
+        WRITE_STAB(sptr, fileidx[0], 0, 0, 0, stabstrlen);
 
         /* this is the stab for the main source file */
         WRITE_STAB(sptr, fileidx[mainfileindex], N_SO, 0, 0, 0);
